/*
   Copyright (c) CNRS,
   @author : Alireza NAKHAEI   anakhaei@laas.fr

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
modification,
   are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
notice,
   this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
   * Neither the name of the <ORGANIZATION> nor the names of its
contributors
   may be used to endorse or promote products derived from this software
without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS
   OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY
   AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER
   OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
   IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef HPP_CELL_H
#define HPP_CELL_H

/**
  \brief Cell is used to discretize 3D environment.

  Each cell stores a set of nodes and edges in which the robot has non-empty intersection with the cell.
 */


/*************************************
INCLUDE
**************************************/
#include <iostream>
#include <fstream>
#include <vector>


#include "KineoWorks2/kwsNode.h"
#include "KineoWorks2/kwsEdge.h"

#include <time.h>
#include <sys/time.h>


#include <set>

using std::set;
using namespace::std;

KIT_PREDEF_CLASS ( ChppCell );

#define DEBUG_C(x)
//#define DEBUG_C(x) std::cerr << "hppCell :" << x << std::endl


/*************************************
CLASS
**************************************/
/**
   \brief Implement the class cell which is a part of environment which include some grids.

   The 3Dgrid envirenmet is discritized to cells. Each cell stores a set of nodes and edges in which the robot has non-empty intersection with the cell.
*/


class ChppCell
{
	public:
		/** \brief Destructor. */
		virtual ~ChppCell() ;

		/** \brief Create a ChppCellShPtr. */
		static ChppCellShPtr create ();

		/** \brief Clear the cell */
		ktStatus clearCell();

		/** \brief Set attupdateFlag */
		ktStatus updateFlag ( bool inUpdateFlag );

		/** \brief Get attupdateFlag */
		bool updateFlag ();

		/** \brief Add a node to the nodeList */
		ktStatus addNode ( CkwsNodeShPtr inNode );

		/** \brief Add a edge to the nodeList */
		ktStatus addEdge ( CkwsEdgeShPtr inEdge );

		/** \brief Return list of nodes */
		std::set<CkwsNodeShPtr> nodeList();

		/** \brief Return list of nodes */
		std::set<CkwsEdgeShPtr> edgeList();

	protected:

		/** \brief Constructor. */
		ChppCell() ;

		/** \brief initialization: store weak pointer to itself */
		ktStatus init ( const ChppCellWkPtr &inWeakPtr );


	private:

		/** \brief Flag which shows wheather the cell has been updated during the last environment update of not */
		bool attUpdateFlag;

		/** \brief Store weak pointer to object.  */
		ChppCellWkPtr attWeakPtr;

		/** \brief A vector includes the nodes which results robot intersection with the cell */
		std::set<CkwsNodeShPtr> attNodeList;

		/** \brief  A vector includes the edges which results robot intersection with the cell   */
		std::set<CkwsEdgeShPtr> attEdgeList;

		/**
		   \if 0
		   \brief Count the number of object created and not destroyed
		*/
		static unsigned int nbObject;
};


#endif
