/*
   Copyright (c) CNRS,
   @author : Alireza NAKHAEI   anakhaei@laas.fr

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
modification,
   are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
notice,
   this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
   * Neither the name of the <ORGANIZATION> nor the names of its
contributors
   may be used to endorse or promote products derived from this software
without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS
   OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY
   AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER
   OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
   IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/


#ifndef hppDynamicEnvHandler_H
#define hppDynamicEnvHandler_H

#include "KineoWorks2/kwsConfig.h"
#include "KineoWorks2/kwsNode.h"
#include "KineoWorks2/kwsEdge.h"

#include "KineoWorks2/kwsDevice.h"

//#include "KineoWorks2/kwsDirectPath.h"
//#include "KineoKCDModel/kppKCDAssembly.h"

#include "kwsPlusRoadmap.h"

//#include "hppModel/hppBody.h"
#include "hppModel/hppDevice.h"


#include "hppDynamicObstacle/hppCell.h"
//#include "hppDynamicObstacle/hppDynamicObstacle.h"

#include <time.h>
#include <sys/time.h>

#define DEF_CELLSIZE	(1)

#define DEF_MAP_AREA_XSTART	0 // [m]
#define DEF_MAP_AREA_YSTART	-1 // [m]
#define DEF_MAP_AREA_ZSTART	1 // [m]
#define DEF_MAP_AREA_XEND	1 // [m]
#define DEF_MAP_AREA_YEND	-1 // [m]
#define DEF_MAP_AREA_ZEND	1 // [m]




KIT_PREDEF_CLASS ( ChppDynamicEnvHandler );
KIT_PREDEF_CLASS ( ChppDynamicObstacle );

//#define DEBUG_DE(x)
#define DEBUG_DE(x) std::cerr << "hppDynamicEnvHandler :" << x << std::endl


typedef struct
{
	double xstart, ystart, zstart;
	double xend, yend,zend;
}
volume3D;

/**
  \brief Main class of package hppDynamicObstacle for implementing roadmap management in changing environment. It discretize the environment to cells. By setting attRoadmap, the roadmap will be modified based on the environment updates. Check the main page for more information regarding how to use the package.
 */

class ChppDynamicEnvHandler
{
	public:

		virtual ~ChppDynamicEnvHandler();

		static ChppDynamicEnvHandlerShPtr create ( ChppDeviceShPtr inDevice );

		/** \brief Set size of Cell [m]. */
		ktStatus setCellSize ( double inCellSize );

		/** \brief Get size of Cell [m]. */
		double getCellSize ();

		/** \brief Allocate the volume needed for the map. we should call the setResolution befor Allocate the map [m].*/
		ktStatus allocateMap ( double xstart, double ystart, double zstart,double xend, double yend, double zend );

		/** \brief Set attRoadmap wich will be modified based on the modification in the environment */
		ktStatus setRoadmap ( CkwsPlusRoadmapShPtr inRoadmap );

		/** \brief Update attRoadmap based on the last modification of the environemnt. */
		ktStatus updateRoadmap();

		/** \brief clear the updateFlag and the lists of all cells and  set the attObstacleIsModifiedFlag of all obstacle to false. */
		ktStatus resetEnvHandler();

		/** \brief Set the assembly of obstacle which will be used in validating nodes and edges of the roadmap */
		ktStatus addObstacle ( ChppDynamicObstacleShPtr inObstacle );

		/** \brief Return a pointer to the cell which include the point 3D */
		ChppCellShPtr posToCell ( double x, double y, double z );

		/** \brief Return the geometry of 3D env */
		volume3D envGeometry();


	protected:

		/** \brief initialization: store weak pointer to itself */
		ktStatus init ( const ChppDynamicEnvHandlerWkPtr &inWeakPtr );

		/** \brief Constructor */
		ChppDynamicEnvHandler ( ChppDeviceShPtr inDevice );



	private:


		/** \brief clear the updateFlag and the lists of all cells. */
		ktStatus clearCellMap();

		/** \brief Remove non-modified obstacle from outer body list  */
		ktStatus setObstacles();

		/** \brief Reset the outer body to the initial state with all modified and non-modified obstacle.  */
		ktStatus reSetObstacles();

		/** \brief Update Cell updateFlag  */
		ktStatus computeModifiedWorkSpaceCells();

		/** \brief Update Cell lists  */
		ktStatus updateCellLists();

		/** \brief Return the cells which the robot has non-empty intersection with in the configuration. */
		std::set<ChppCellShPtr> configToCells ( CkwsConfig inConfig );

		/** \brief Add node to the nodelist of concerned cells */
		ktStatus addNode ( CkwsNodeShPtr inNode );

		/** \brief Add edge to the edgelist of concerned cells */
		ktStatus addEdge ( CkwsEdgeShPtr inEdge );

		/** \brief Return true if Node is valid. */
		bool isNodeValid ( CkwsNodeShPtr inNode );

		/** \brief Return true if Edge is valid. */
		bool isEdgeValid ( CkwsEdgeShPtr inEdge );

		/** \brief Return the status of each cell. */
		ktStatus statusTest();

		/** \brief Store weak pointer to object.  */
		ChppDynamicEnvHandlerWkPtr attWeakPtr;

		/** \brief Cell size [m] */
		double attCellSize;

		/** \brief The discritize map */
		std::vector<ChppCellShPtr> attCellMap;

		/** \brief Geometry of the 3D environmet */
		volume3D att3DMapArea;

		/** \brief Roadmap */
		CkwsPlusRoadmapShPtr attRoadmap;

		/** \brief Obstacle list */
		std::set<ChppDynamicObstacleShPtr> attObstacleList;

		/** \brief Robot */
		ChppDeviceShPtr attDevice;

		/** \brief The number of cell in the map */
		int attNbOfCell;

		/** \brief The number of cell in the map along X */
		int attNbCellX;

		/** \brief The number of cell in the map along Y */
		int attNbCellY;

		/** \brief The number of cell in the map along Z */
		int attNbCellZ;

		/** \brief Pointer to directPath validator */
		CkwsValidatorDPCollisionShPtr attDirectPathvalidator;

		/** \brief Pointer to configure validator */
		CkwsConfigValidatorSetConstShPtr attConfigValidator;

		/** \brief Map between each body and its assigned outer body */
		std::map<CkwsKCDBodyShPtr, std::vector<CkcdObjectShPtr> > attOuterBodyMap;

		/**
		   \if 0
		   \brief Count the number of object created and not destroyed
		*/
		static unsigned int nbObject;
};
#endif
