/*
   Copyright (c) CNRS,
   @author : Alireza NAKHAEI   anakhaei@laas.fr

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
modification,
   are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
notice,
   this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
   * Neither the name of the <ORGANIZATION> nor the names of its
contributors
   may be used to endorse or promote products derived from this software
without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS
   OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY
   AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER
   OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
   IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/


#ifndef hpp_polyhedron_map_H
#define hpp_polyhedron_map_H

#include <iostream>
#include <fstream>
#include <vector>
#include <set>


#include <time.h>
#include <sys/time.h>

#include "KineoKCDModel/kppKCDBox.h"
#include "KineoKCDModel/kppKCDPolyhedron.h"
#include "KineoKCDModel/kppKCDAssembly.h"

#define X_MAP 400
#define Y_MAP 400
#define Z_MAP 100

using std::set;
using namespace::std;

KIT_PREDEF_CLASS ( ChppPolyhedronMap );

#define DEBUG_MAP(x)
//#define DEBUG_MAP(x) std::cerr << "ChppPolyhedronMap :" << x << std::endl


/**
  \brief class of package slam3DPlanner. It create a map by adding or removing kppKCDBox object to an assembly.
 */

class ChppPolyhedronMap
{
	public:
		/** \brief Destructor. */
		~ChppPolyhedronMap();

		/** \brief Create a ChppPolyhedronMapShPtr. */
		static ChppPolyhedronMapShPtr create ( const std::string &i_name , double inSize );

		/** \brief Add a kppKCDBox to the map. */
		ktStatus addPolyhedron ( int i, int j, int k );

		/** \brief Add a kppKCDBox to the map. */
		ktStatus addPolyhedron ( double x, double y, double z );

		/** \brief Remove a kppKCDBox from the. */
		ktStatus removePolyhedron ( int i, int j, int k );

		/** \brief Remove a kppKCDBox from the. */
		ktStatus removePolyhedron ( double x, double y, double z );

		/** \brief Return a pointer to the assembly of obstacle. */
		CkppKCDAssemblyShPtr getPolyhedronMap();

		/** \brief Reset UpdateReport. */
		std::vector<double> idToPos ( int id );

		/** \brief Get map Resolution. */
		double getSize();

		/** \brief Sets the diffuse color of the map */
		void diffuseColor (CkppColor inColor );

		/** \brief Clear the map assembly */
		void clear( );

	protected:

		/** \brief Constructor. */
		ChppPolyhedronMap ( const std::string &i_name, double inSize );

		/** \brief initialization: store weak pointer to itself */
		ktStatus init ( const ChppPolyhedronMapWkPtr &inWeakPtr );


	private:

		/** \brief Return xyz from ijk. */
		std::vector<double> ijkToPos ( int i, int j, int k );

		/** \brief Return Box number from its position. */
		int posToID ( double x, double y, double z );

		/** \brief Return a CkppKCDBoxShPtr at the position */
		CkppKCDBoxShPtr createBox ( const std::string &inName, const double x, const double y, const double z );

		/** \brief Store weak pointer to object.  */
		ChppPolyhedronMapWkPtr attWeakPtr;

		/** \brief Vector of created CkppKCDPolyhedron */
		std::vector<CkppKCDBoxShPtr> attBoxVector;

		/** \brief Vector of created CkppKCDPolyhedron */
		double attResolution;

		/** \brief PolyhedronMap */
		CkppKCDAssemblyShPtr attPolydedronMap;

		/** \brief Color */
		CkppColor attColor;

		/** \brief Assembly name */
		std::string attAssemblyName;


		/**
		   \if 0
		   \brief Count the number of object created and not destroyed
		*/
		static unsigned int nbObject;
};

#endif
