/*
   Copyright (c) CNRS,
   @author : Alireza NAKHAEI   anakhaei@laas.fr

   All rights reserved.

   Redistribution and use in source and binary forms, with or without
modification,
   are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
notice,
   this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
   * Neither the name of the <ORGANIZATION> nor the names of its
contributors
   may be used to endorse or promote products derived from this software
without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS
   OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY
   AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER
   OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
   IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

/*****************************************
 INCLUDES
*******************************************/

#include <iostream>
#include "hppDynamicObstacle/hppDynamicEnvHandler.h"
#include "hppDynamicObstacle/hppDynamicObstacle.h"

// Select verbosity at configuration by setting CXXFLAGS="... -DDEBUG=[1 or 2]"
#if DEBUG==2
#define ODEBUG2(x) std::cout << "hppDynamicEnvHandler:" << x << std::endl
#define ODEBUG1(x) std::cerr << "hppDynamicEnvHandler:" << x << std::endl
#elif DEBUG==1
#define ODEBUG2(x)
#define ODEBUG1(x) std::cerr << "hppDynamicEnvHandler:" << x << std::endl
#else
#define ODEBUG2(x)
#define ODEBUG1(x)
#endif


unsigned int ChppDynamicEnvHandler::nbObject = 0;

// ==========================================================================
ChppDynamicEnvHandler::ChppDynamicEnvHandler ( ChppDeviceShPtr inDevice )
{
	attDevice= inDevice;

	attDirectPathvalidator=CkwsValidatorDPCollision::create ( attDevice,CkwsValidatorDPCollision::PROGRESSIVE_BACKWARD );
	attDirectPathvalidator->isActivated ( true );
	attDirectPathvalidator->penetration ( 0.02 );

	attConfigValidator=attDevice->configValidators();

	nbObject++;
	ODEBUG2 ( "Constructor of ChppDynamicEnvHandler. number of ChppDynamicEnvHandler objects"<< nbObject );

}
// ==========================================================================
ChppDynamicEnvHandler::~ChppDynamicEnvHandler()
{
	nbObject--;
	ODEBUG2 ( "--------------------------------------------------------------------------------------------" );
	ODEBUG2 ( "      Destructor of class ChppDynamicEnvHandler. number of ChppDynamicEnvHandler objects:" << nbObject );
	ODEBUG2 ( "--------------------------------------------------------------------------------------------" );
}
// ==========================================================================
ChppDynamicEnvHandlerShPtr ChppDynamicEnvHandler::create ( ChppDeviceShPtr inDevice )
{
	ChppDynamicEnvHandler* flatPtr = new ChppDynamicEnvHandler ( inDevice );
	ChppDynamicEnvHandlerShPtr flatShPtr ( flatPtr );
	ChppDynamicEnvHandlerWkPtr flatWkPtr ( flatShPtr );

	ODEBUG2 ( "ChppDynamicEnvHandler create" );
	if ( flatPtr->init ( flatWkPtr ) != KD_OK ) flatShPtr.reset();

	return flatShPtr;
}
// ==========================================================================
ktStatus ChppDynamicEnvHandler::init ( const ChppDynamicEnvHandlerWkPtr &inWeakPtr )
{
	attWeakPtr = inWeakPtr;

	return KD_OK;
}
// ==========================================================================
ktStatus ChppDynamicEnvHandler::setCellSize ( double inCellSize )
{
	attCellSize= inCellSize;
	return KD_OK;
}
// ==========================================================================
double ChppDynamicEnvHandler::getCellSize ()
{
	return attCellSize;
}
// ==========================================================================
ktStatus ChppDynamicEnvHandler::allocateMap ( double xstart, double ystart, double zstart,double xend, double yend, double zend )
{
	ODEBUG2 ( "Allocating map ....." );
	att3DMapArea.xstart=xstart;
	att3DMapArea.ystart=ystart;
	att3DMapArea.zstart=zstart;
	att3DMapArea.xend=xend;
	att3DMapArea.yend=yend;
	att3DMapArea.zend=zend;

	attNbCellX = ( int ) ceil ( ( att3DMapArea.xend-att3DMapArea.xstart ) /attCellSize -0.00001 );
	attNbCellY = ( int ) ceil ( ( att3DMapArea.yend-att3DMapArea.ystart ) /attCellSize -0.00001 );
	attNbCellZ = ( int ) ceil ( ( att3DMapArea.zend-att3DMapArea.zstart ) /attCellSize -0.00001 );

	attNbOfCell=attNbCellX *attNbCellY*attNbCellZ;

	ODEBUG2 ( "NbCellX =  " << attNbCellX << " ,NbCellY =  " << attNbCellY << " , NbCellZ =  " << attNbCellZ );

	attCellMap.resize ( attNbOfCell );

	for ( int i=0; i<attNbOfCell ; i++ )
	{
		attCellMap[i] = ChppCell::create();
	}

	ODEBUG2 ( "size of cellMap =  " << attNbOfCell );

	if ( resetEnvHandler() !=KD_OK )
	{
		cout << "ERROR : ChppDynamicObstacle::allocateMap - error in clearMap " <<endl;
		return KD_ERROR;
	}

	ODEBUG2 ( "....map was allocated." );
	return KD_OK;
}

// ==========================================================================
ktStatus ChppDynamicEnvHandler::setRoadmap ( CkwsPlusRoadmapShPtr inRoadmap )
{
	attRoadmap=inRoadmap;
	return KD_OK;
}

// ==========================================================================
ktStatus ChppDynamicEnvHandler::updateRoadmap()
{
	ODEBUG2 ( "Updating Roadmap ....." );

	if ( clearCellMap() !=KD_OK )
	{
		cout << "ERROR : ChppDynamicObstacle::updateRoadmap() - error in clearMap " <<endl;
		return KD_ERROR;
	}
	ODEBUG2 ( "Cell report after clearing Cell Map: " );

	computeModifiedWorkSpaceCells();
	updateCellLists();
	std::set<CkwsNodeShPtr> nodeList;
	std::set<CkwsNodeShPtr> notValidNodeList;
	std::set <CkwsNodeShPtr>:: iterator nodeIt;

	std::set<CkwsEdgeShPtr> edgeList;
	std::set<CkwsEdgeShPtr> notValidEdgeList;
	std::set <CkwsEdgeShPtr>:: iterator edgeIt;

	for ( int i=0; i<attNbOfCell;i++ )
	{
		if ( attCellMap.at ( i )->updateFlag() == true )
		{
			//getting nodes
			std::set<CkwsNodeShPtr> localnodeList=attCellMap.at ( i )->nodeList();
			for ( nodeIt=localnodeList.begin();nodeIt != localnodeList.end();nodeIt++ )
				nodeList.insert ( *nodeIt );

			//getting Edges
			std::set<CkwsEdgeShPtr> localEdgeList=attCellMap.at ( i )->edgeList();
			for ( edgeIt=localEdgeList.begin();edgeIt != localEdgeList.end();edgeIt++ )
				edgeList.insert ( *edgeIt );
		}
	}

// Clearing outer body list from non-modified obstacle
	if ( setObstacles() !=KD_OK )
	{
		ODEBUG1 ( "ERROR : ChppDynamicObstacle::updateRoadmap - error in setObstacles " );
		return KD_ERROR;
	}

	ODEBUG2 ( "Number of NODES for test=  " << nodeList.size() );
	for ( nodeIt=nodeList.begin();nodeIt != nodeList.end();nodeIt++ )
		if ( !isNodeValid ( *nodeIt ) )
		{
			notValidNodeList.insert ( *nodeIt );
			ODEBUG2 ( "updateRoadmap():Node is not valid" );
		}

	ODEBUG2 ( "Number of EDGES for test =  " << edgeList.size() );
	for ( edgeIt=edgeList.begin();edgeIt != edgeList.end();edgeIt++ )
		if ( !isEdgeValid ( *edgeIt ) )
		{
			notValidEdgeList.insert ( *edgeIt );
			ODEBUG2 ( "updateRoadmap():Edge is not valid" );
		}

// Resetting outer body list
	if ( reSetObstacles() !=KD_OK )
	{
		ODEBUG1 ( "ERROR : ChppDynamicObstacle::updateRoadmap - error in reSetObstacles " );
		return KD_ERROR;
	}

	ODEBUG2 ( "Number of not valid NODES =  " << notValidNodeList.size() );
	ODEBUG2 ( "Number of not valid EDGES =  " << notValidEdgeList.size() );

	if ( attRoadmap->removeNodeEdge ( notValidNodeList,notValidEdgeList ) !=KD_OK )
	{
		ODEBUG2 ( "ERROR in removing node and edge" );
		return KD_ERROR;
	}

// clear cell's lists and clear obtacle flag
	resetEnvHandler();

	ODEBUG2 ( "....Roadmap was Updated." );

	return KD_OK;
}
// ==========================================================================
ktStatus ChppDynamicEnvHandler::addObstacle ( ChppDynamicObstacleShPtr inObstacle )
{
	ODEBUG2 ( "Adding obstacle ...." );

	attObstacleList.insert ( inObstacle );

	ODEBUG2 ( "....Obstacle is added." );
	return KD_OK;
}
// ==========================================================================
ktStatus ChppDynamicEnvHandler::computeModifiedWorkSpaceCells()
{
	ODEBUG2 ( "Computing modified Cell....." );
	std::set <ChppDynamicObstacleShPtr>::iterator obstacleIt;
	int i=0;

	for ( obstacleIt=attObstacleList.begin();obstacleIt != attObstacleList.end();obstacleIt++ )
	{
		ODEBUG2 ( "test = " << i );
		if ( ( *obstacleIt )->obstacleModificationFlag() == true )
		{
			( *obstacleIt )->computeModifiedWorkSpaceCells();
		}
		i++;
	}

	ODEBUG2 ( "....Modified cells were computed." );
	return KD_OK;
}

// ==========================================================================
ChppCellShPtr ChppDynamicEnvHandler::posToCell ( double x, double y, double z )
{

	if ( x<att3DMapArea.xstart )
		x=att3DMapArea.xstart+.00001;

	if ( y<att3DMapArea.ystart )
		y=att3DMapArea.ystart+.00001;

	if ( z<att3DMapArea.zstart )
		z=att3DMapArea.zstart+.00001;

	if ( x>att3DMapArea.xend )
		x=att3DMapArea.xend-.00001;

	if ( y>att3DMapArea.yend )
		y=att3DMapArea.yend-.00001;

	if ( z>att3DMapArea.zend )
		z=att3DMapArea.zend-.00001;

	int xx= ( int ) floor ( ( x-att3DMapArea.xstart ) / ( attCellSize ) + 0.0001 );
	int yy= ( int ) floor ( ( y-att3DMapArea.ystart ) / ( attCellSize ) + 0.0001 );
	int zz= ( int ) floor ( ( z-att3DMapArea.zstart ) / ( attCellSize ) + 0.0001 );

	if ( xx==attNbCellX )
		xx=xx-1;
	if ( yy==attNbCellY )
		yy=yy-1;
	if ( zz==attNbCellZ )
		zz=zz-1;

	return attCellMap.at ( xx+yy*attNbCellX + zz*attNbCellX*attNbCellY );
}
// ==========================================================================
std::set<ChppCellShPtr> ChppDynamicEnvHandler::configToCells ( CkwsConfig inConfig )
{
	std::set<ChppCellShPtr> cellList;
	double xMin, yMin, zMin, xMax, yMax, zMax;

	if ( ! ( attDevice-> hppSetCurrentConfig ( inConfig, ChppDevice::GEOMETRIC ) ) )
	{
		ODEBUG1 ( "ERROR : Cslam3DPlanner::updateCellLists: error in setting the robot in desired config." );
	}
	if ( ( attDevice-> axisAlignedBoundingBox ( xMin, yMin, zMin, xMax, yMax, zMax ) ) != KD_OK )
	{
		ODEBUG1 ( "ERROR : Cslam3DPlanner::updateCellLists: error in geting the dimention of bounding box." );
	}
// To detect the cells which are in collison with the robot
	std::vector<double> xVec;
	std::vector<double> yVec;
	std::vector<double> zVec;

	for ( double i=0 ; i<=1 ; i=i+1 )
		xVec.push_back ( xMin+i* ( xMax-xMin ) );

	for ( double i=0 ; i<=1 ; i=i+1 )
		yVec.push_back ( yMin+i* ( yMax-yMin ) );

	for ( double i=0 ; i<=1 ; i=i+0.5 )
		zVec.push_back ( zMin+i* ( zMax-zMin ) );

	for ( unsigned int k=0; k<zVec.size();k++ )
	{
		for ( unsigned int j=0; j<yVec.size();j++ )
		{
			for ( unsigned int i=0; i<xVec.size();i++ )
			{
				if ( ( xVec.at ( i ) > att3DMapArea.xstart ) && ( xVec.at ( i ) < att3DMapArea.xend ) && ( yVec.at ( j ) >att3DMapArea.ystart ) && ( yVec.at ( j ) < att3DMapArea.yend ) && ( zVec.at ( k ) > att3DMapArea.zstart ) && ( zVec.at ( k ) < att3DMapArea.zend ) )
				{
					cellList.insert ( posToCell ( xVec.at ( i ),yVec.at ( j ),zVec.at ( k ) ) );
				}
			}
		}
	}

	return cellList;
}
// ==========================================================================
ktStatus ChppDynamicEnvHandler::addNode ( CkwsNodeShPtr inNode )
{
	std::set<ChppCellShPtr> cellList=configToCells ( inNode->config() );
	std::set <ChppCellShPtr>::iterator cellIt;

	for ( cellIt=cellList.begin();cellIt != cellList.end();cellIt++ )
		( *cellIt )->addNode ( inNode );
	return KD_OK;
}
// ==========================================================================
ktStatus ChppDynamicEnvHandler::addEdge ( CkwsEdgeShPtr inEdge )
{
	std::set<ChppCellShPtr> cellList;
	std::set<ChppCellShPtr>::iterator cellListIt;

	CkwsNodeShPtr endNode=inEdge->endNode ( );
	CkwsConfig tempConf = endNode->config() ;

	std::set<ChppCellShPtr> cellListLocal=configToCells ( tempConf );

	for ( cellListIt=cellListLocal.begin();cellListIt != cellListLocal.end();cellListIt++ )
		cellList.insert ( *cellListIt );


	CkwsDirectPathConstShPtr directPath=inEdge->directPath ();
	double pathLength = directPath->length();

	double u=0;

	while ( u<=pathLength )
	{
		directPath->getConfigAtDistance ( u, tempConf );
		cellListLocal=configToCells ( tempConf );

		for ( cellListIt=cellListLocal.begin();cellListIt != cellListLocal.end();cellListIt++ )
			cellList.insert ( *cellListIt );

		u=u+attCellSize/20;
	}

	for ( cellListIt=cellList.begin();cellListIt != cellList.end();cellListIt++ )
		( *cellListIt )->addEdge ( inEdge );

	return KD_OK;
}
// ==========================================================================
bool ChppDynamicEnvHandler::isNodeValid ( CkwsNodeShPtr inNode )
{
	ODEBUG2 ( "isNodeValid.....  " );

	CkwsDeviceShPtr device=inNode->config().device();
	if ( !device )
	{
		ODEBUG1 ( "isNodeValid : ERROR : NO Device " );
	}

	if ( device->configValidators()->validate ( inNode->config() ) )
	{
		return true;
	}
	else
	{
		ODEBUG2 ( ": isNodeValidNode: Node in not valid  " );
		return false;
	}
	/*	CkwsConfig config=inNode->config();

		if ( ! ( attConfigValidator->validate ( config ) ) )
		{
			ODEBUG1 ( " Not SUUCCEFUL IN VALIDATING THE NODE " );
		}

		bool confIsValid=config.isValid();
		if ( !confIsValid )
		{
			ODEBUG2 ( "Node in not valid  " );
		}
		return confIsValid;*/
}
// ==========================================================================
bool ChppDynamicEnvHandler::isEdgeValid ( CkwsEdgeShPtr inEdge )
{
	ODEBUG2 ( "isEdgeValid ....  " );
	CkwsDeviceShPtr device=inEdge->endNode()->config().device();
	if ( !device )
		ODEBUG1 ( "isEdgeValid : ERROR : NO Device " );


	CkwsDirectPathValidatorSetConstShPtr directPathValidator = device->directPathValidators	( );
	if ( !directPathValidator )
		ODEBUG1 ( " isEdgeValid ::Error: directPathValidator is NULL " );


	if ( directPathValidator->validate ( * ( inEdge-> directPath() ) ) )
		return true;
	else
	{
		ODEBUG2 ( ": isEdgeValid: Edge in not valid  " );
		return false;
	}
/*

	CkwsDirectPathConstShPtr directPath = inEdge-> directPath () ;
	if ( !directPath )
		ODEBUG1 ( "ERROR :  directPath " );

	bool pathIsValid = attDirectPathvalidator->validate ( * ( directPath.get() ) );

	if ( !pathIsValid )
	{
		ODEBUG2 ( "Edge is not valid  " );
	}
	return pathIsValid;*/
}
// ==========================================================================
ktStatus ChppDynamicEnvHandler::resetEnvHandler()
{
	ODEBUG2 ( "resetEnvHandler....." );

	clearCellMap();

	std::set<ChppDynamicObstacleShPtr>:: iterator obstacleListIt;
	for ( obstacleListIt=attObstacleList.begin();obstacleListIt != attObstacleList.end();obstacleListIt++ )
		( *obstacleListIt )->roadmapIsModified();

	ODEBUG2 ( " ....resetEnvHandler" );
	return KD_OK;
}

// ==========================================================================
ktStatus ChppDynamicEnvHandler::clearCellMap()
{
	ODEBUG2 ( "clearCellMap....." );

	for ( int i=0; i<attNbOfCell; i++ )
	{
		if ( ! attCellMap.at ( i ) )
			ODEBUG1 ( "ERROR - NO Cell " << i );
		attCellMap.at ( i )->clearCell();
	}
	ODEBUG2 ( " ....clearCellMap" );
	return KD_OK;
}

//=========================================================================================
ktStatus ChppDynamicEnvHandler::updateCellLists()
{
	//gettimeofday ( Tps, Tzp );
	ODEBUG2 ( "Updating cell lists....." );
	int nbOfNodeInRoadMap=attRoadmap->countNodes ();
	unsigned int nbOfEdgeInRoadMap=0;

	for ( int i =0 ; i< nbOfNodeInRoadMap ; i++ )
	{
		CkwsNodeShPtr inNode=attRoadmap->node ( i ) ;
		addNode ( inNode );

		unsigned int nbOutEdge = inNode->countOutEdges() ;

		for ( unsigned int j=0 ; j< nbOutEdge ; j++ )
		{
			addEdge ( inNode->outEdge ( j ) );
		}

		nbOfEdgeInRoadMap = nbOfEdgeInRoadMap+ nbOutEdge ;
	}

	ODEBUG2 ( "NUMBER OF NODES IN ROADMAP =  " << nbOfNodeInRoadMap );
	ODEBUG2 ( "NUMBER OF EDGES IN ROADMAP =  " << nbOfEdgeInRoadMap );
	ODEBUG2 ( "....cell lists are Updated." );
	return KD_OK;
}
//=========================================================================================
ktStatus ChppDynamicEnvHandler::setObstacles()
{
	ODEBUG2 ( "setObstacle : set just the modified obstacle in the outer bject list...." );

	CkwsDevice::TBodyVector bodyVector;
	attDevice->getBodyVector ( bodyVector );

	int vectorSize=bodyVector.size();

	for ( int i=0; i< vectorSize; i++ )
	{
		CkwsKCDBodyShPtr kcdBody=KIT_DYNAMIC_PTR_CAST ( CkwsKCDBody, bodyVector[i] );

		std::vector<CkcdObjectShPtr> outerObjs = kcdBody->outerObjects();

		attOuterBodyMap[kcdBody]=outerObjs;


		std::vector< CkcdObjectShPtr > i_outerObjects;

		std::set<ChppDynamicObstacleShPtr>:: iterator obstacleListIt;
		for ( obstacleListIt=attObstacleList.begin();obstacleListIt != attObstacleList.end();obstacleListIt++ )
		{

			if ( ( *obstacleListIt )->obstacleModificationFlag() == true )
			{
				i_outerObjects.push_back ( ( *obstacleListIt )->getModifiedObstacle() );
			}
		}
		kcdBody->outerObjects ( i_outerObjects );
	}
	ODEBUG2 ( ".......setObstacle." );

	return KD_OK;
}
//=========================================================================================
ktStatus ChppDynamicEnvHandler::reSetObstacles()
{
	ODEBUG2 ( "reSetObstacle : ...." );

	CkwsDevice::TBodyVector bodyVector;
	attDevice->getBodyVector ( bodyVector );

	int vectorSize=bodyVector.size();

	for ( int i=0; i< vectorSize; i++ )
	{
		CkwsKCDBodyShPtr kcdBody=KIT_DYNAMIC_PTR_CAST ( CkwsKCDBody, bodyVector[i] );

		kcdBody->outerObjects ( attOuterBodyMap[kcdBody] );
	}
	attOuterBodyMap.clear();
	ODEBUG2 ( ".......reSetObstacle." );

	return KD_OK;
}
//=========================================================================================
volume3D ChppDynamicEnvHandler::envGeometry()
{
	return att3DMapArea;
}
// ============================================================================
ktStatus ChppDynamicEnvHandler::statusTest()
{
	cout << "ChppDynamicEnvHandler::Cells report:  " << endl;

	for ( int i=0 ; i<attNbOfCell ; i++ )
	{
		int nodeVecSize = attCellMap.at ( i )->nodeList().size();
		int edgeVecSize = attCellMap.at ( i )->edgeList().size();
		cout <<"Cell " << i << " : Number of Node-> " <<nodeVecSize << " , Number of Edge-> " <<edgeVecSize << ",  Update Flag -> " <<attCellMap.at ( i )->updateFlag() << endl;
	}
	return KD_OK;
}
